//
// Cryptography
//
// Copyright 2005 John Reilly
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Linking this library statically or dynamically with other modules is
// making a combined work based on this library.  Thus, the terms and
// conditions of the GNU General Public License cover the whole
// combination.
// 
// As a special exception, the copyright holders of this library give you
// permission to link this library with independent modules to produce an
// executable, regardless of the license terms of these independent
// modules, and to copy and distribute the resulting executable under
// terms of your choice, provided that you also meet, for each linked
// independent module, the terms and conditions of the license of that
// module.  An independent module is a module which is not derived from
// or based on this library.  If you modify this library, you may extend
// this exception to your version of the library, but you are not
// obligated to do so.  If you do not wish to do so, delete this
// exception statement from your version.
//

using System;
using System.IO;

namespace ICSharpCode.SharpZipLib.Encryption
{

#if COMPACT_FRAMEWORK

	public class CryptographicException : SystemException
	{
		public CryptographicException() : base()
		{
			// Do nothing
		}
		
		public CryptographicException(string message) : base(message)
		{
			// Do nothing
		}
	}
	
	public interface ICryptoTransform : IDisposable
	{
		int TransformBlock(byte[] inputBuffer, int inputOffset, int inputCount, byte[] outputBuffer, int outputOffset);
		byte[] TransformFinalBlock(byte[] inputBuffer, int inputOffset, int inputCount);
		
		bool CanReuseTransform { get; }
		bool CanTransformMultipleBlocks { get; }
		int InputBlockSize { get; }
		int OutputBlockSize { get; }
	}
	
	public sealed class KeySizes
	{
		public KeySizes(int minSize, int maxSize, int skipSize)
		{
			minSize_ = minSize;
			maxSize_ = maxSize;
			skipSize_ = skipSize;
		}
	
		public int MaxSize
		{
			get { return maxSize_; }
		}
		
		public int MinSize 
		{ 
			get { return minSize_; }
		}
		
		public int SkipSize
		{ 
			get { return skipSize_; }
		}
	
		int maxSize_;
		int minSize_;
		int skipSize_;
	}
	
	public abstract class SymmetricAlgorithm : IDisposable
	{
		public SymmetricAlgorithm()
		{
			// Do nothing
		}

		~SymmetricAlgorithm()
		{
			Dispose(false);
		}

		public void Clear()
		{
			((IDisposable)this).Dispose();
		}
		
		public static SymmetricAlgorithm Create()
		{
			return null;
		}

		public static SymmetricAlgorithm Create(string algName)
		{
			return null;
		}
		
		public virtual ICryptoTransform CreateDecryptor()
		{
			return this.CreateDecryptor(Key, IV);
		}
		
		public abstract ICryptoTransform CreateDecryptor(byte[] rgbKey, byte[] rgbIV);
		
		public virtual ICryptoTransform CreateEncryptor()
		{
			return CreateEncryptor(Key, IV);
		}
		
		public abstract ICryptoTransform CreateEncryptor(byte[] rgbKey, byte[] rgbIV);
		
		protected virtual void Dispose(bool disposing)
		{
			if (disposing)
			{
				if ( KeyValue != null ) 
				{
					Array.Clear(KeyValue, 0, KeyValue.Length);
					KeyValue = null;
				}
				if ( IVValue != null ) 
				{
					Array.Clear(IVValue, 0, IVValue.Length);
					IVValue = null;
				}
			}
		}

		void IDisposable.Dispose()
		{
			Dispose(true);
		}
		
		public abstract void GenerateIV();
		
		public abstract void GenerateKey();
		
		public bool ValidKeySize(int bitLength)
		{
			return true;
		}
		
		public virtual KeySizes[] LegalKeySizes
		{
			get 
			{
				return (KeySizes[]) LegalKeySizesValue.Clone();
			}
		}
		
		public virtual KeySizes[] LegalBlockSizes
		{
			get 
			{
				return (KeySizes[]) LegalBlockSizesValue.Clone();
			}
		}

		public int KeySize
		{
			get 
			{
				return KeySizeValue;
			}
			
			set 
			{
				if ( !ValidKeySize(value) ) 
				{
					throw new CryptographicException("Invalid key size");
				}
				KeySizeValue = value;
				KeyValue = null;
			}
		}
		
		public virtual int BlockSize
		{
			get 
			{
				return BlockSizeValue;
			}
			
			set 
			{
				for (int index = 0; index < LegalBlockSizesValue.Length; ++index)
				{
					if (LegalBlockSizesValue[index].SkipSize == 0)
					{
						if (LegalBlockSizesValue[index].MinSize != value)
						{
							continue;
						}
						BlockSizeValue = value;
						IVValue = null;
						return;
					}

					for (int size = this.LegalBlockSizesValue[index].MinSize; size <= this.LegalBlockSizesValue[index].MaxSize; size += this.LegalBlockSizesValue[index].SkipSize)
					{
						if (size == value)
						{
							if ( BlockSizeValue != value )
							{
								BlockSizeValue = value;
								IVValue = null;
							}
							return;
						}
					}
				}
				throw new CryptographicException("Block size is invalid");
			}
		}
		
		public virtual byte[] IV 
		{ 
			get 
			{
				if (IVValue == null) 
				{
					GenerateIV();
				}
				return (byte[]) IVValue.Clone();
			}
			
			set 
			{
				if (value == null) 
				{
					throw new ArgumentNullException("value");
				}
				if (value.Length > (BlockSizeValue / 8)) 
				{
					throw new CryptographicException("Invalid IV size");
				}
				IVValue = (byte[]) value.Clone();
			}
		}
	
		public abstract byte[] Key { get; set; }
	
		protected byte[] IVValue;
		protected byte[] KeyValue;
		protected int KeySizeValue;
		protected int BlockSizeValue;
		protected KeySizes[] LegalBlockSizesValue;
		protected KeySizes[] LegalKeySizesValue;
		protected int FeedbackSizeValue;
	}
	
	public enum CryptoStreamMode
	{
		Read,
			Write
		}
	
	public class CryptoStream : Stream, IDisposable
	{
		public CryptoStream(Stream stream, ICryptoTransform transform, CryptoStreamMode mode)
		{
			stream_ = stream;
			transform_ = transform;
			mode_ = mode;
			finalBlockTransformed_ = false;
			canWrite_ = false;

			// Throw exception here if stream_ cannot write?
			if ( mode_ == CryptoStreamMode.Write ) 
			{
				canWrite_ = stream_.CanWrite;
			}
			
			canRead_ = false;

			// Throw exception here if stream_ cannot read?
			if ( mode == CryptoStreamMode.Read ) 
			{
				canRead_ = stream_.CanRead;
			}
		}

		~CryptoStream()
		{
			Dispose(false);
		}
		
		public void Clear()
		{
			
		}
		public override void Close()
		{
			// Do nothing???		
		}

		protected virtual void Dispose(bool disposing)
		{
			if ( disposing ) 
			{
				// ??
			}
		}

		void IDisposable.Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}
		
		public override void Flush()
		{
			// Do nothing			
		}
		
		public void FlushFinalBlock()
		{
			if ( finalBlockTransformed_ ) 
			{
				throw new CryptographicException("Final block already flushed");
			}
			
			finalBlockTransformed_ = true;
			byte[] buffer = transform_.TransformFinalBlock(InputBuffer_, 0, InputBufferIndex_);
			
			// Flush the block
			if (stream_ is CryptoStream) 
			{
				((CryptoStream) stream_).FlushFinalBlock();
			}
			else 
			{
				stream_.Flush();
			}
		}
		
		private void InitializeBuffer()
		{
			if (transform_ != null)
			{
				InputBlockSize_ = transform_.InputBlockSize;
				InputBuffer_ = new byte[InputBlockSize_];
				OutputBlockSize_ = transform_.OutputBlockSize;
				OutputBuffer_ = new byte[OutputBlockSize_];
			}
			
			if (canRead_)
			{
				NextInputBuffer_ = new byte[InputBlockSize_];
			}
		}
		
		public override int Read(byte[] buffer, int offset, int count)
		{
			int num4;
			byte[] buffer3;
			if (!canRead_)
			{
				throw new NotSupportedException("Cannot read from stream");
			}
			if (offset < 0)
			{
				throw new ArgumentOutOfRangeException("offset", "offset < 0");
			}
			if (count < 0)
			{
				throw new ArgumentOutOfRangeException("count", "Count < 0");
			}
			
			if ((buffer.Length - offset) < count)
			{
				throw new ArgumentException("Offset + length is invalid");
			}

			int localCount = count;
			int localOffset = offset;
			
			if (OutputBufferIndex_ != 0)
			{
				if (OutputBufferIndex_ <= count)
				{
					Buffer.BlockCopy(OutputBuffer_, 0, buffer, offset, OutputBufferIndex_);
					localCount -= OutputBufferIndex_;
					localOffset += OutputBufferIndex_;
					OutputBufferIndex_ = 0;
				}
				else
				{
					Buffer.BlockCopy(OutputBuffer_, 0, buffer, offset, count);
					Buffer.BlockCopy(OutputBuffer_, count, OutputBuffer_, 0, OutputBufferIndex_ - count);
					OutputBufferIndex_ -= count;
					return count;
				}
			}
			
			if (finalBlockTransformed_) 
			{
				return (count - localCount);
			}
			
			int num3 = 0;
			
			if ((localCount > OutputBlockSize_) && transform_.CanTransformMultipleBlocks) 
			{
				int num5 = localCount / OutputBlockSize_;
				int num6 = num5 * InputBlockSize_;
				byte[] buffer1 = new byte[num6];
				Buffer.BlockCopy(InputBuffer_, 0, buffer1, 0, InputBufferIndex_);
				num3 = InputBufferIndex_ + stream_.Read(buffer1, InputBufferIndex_, num6 - InputBufferIndex_);
				InputBufferIndex_ = 0;
				if (num3 <= InputBlockSize_)
				{
					InputBuffer_ = buffer1;
					InputBufferIndex_ = num3;
				}
				else
				{
					int num7 = (num3 / InputBlockSize_) * InputBlockSize_;
					int num8 = num3 - num7;
					if (num8 != 0)
					{
						InputBufferIndex_ = num8;
						Buffer.BlockCopy(buffer1, num7, InputBuffer_, 0, num8);
					}
					byte[] buffer2 = new byte[(num7 / InputBlockSize_) * OutputBlockSize_];
					num4 = transform_.TransformBlock(buffer1, 0, num7, buffer2, 0);
					Buffer.BlockCopy(buffer2, 0, buffer, localOffset, num4);
					Array.Clear(buffer1, 0, buffer1.Length);
					Array.Clear(buffer2, 0, buffer2.Length);
					localCount -= num4;
					localOffset += num4;
				}
			}

			while (InputBufferIndex_ < InputBlockSize_)
			{
				num3 = stream_.Read(InputBuffer_, InputBufferIndex_, InputBlockSize_ - InputBufferIndex_);
				if (num3 == 0)
				{
					goto Label_0332;
				}
				InputBufferIndex_ += num3;
			}

			while (localCount > 0)
			{
				NextInputBufferIndex_ = stream_.Read(NextInputBuffer_, 0, InputBlockSize_);
				if (NextInputBufferIndex_ == 0)
				{
					goto Label_0332;
				}
				num4 = transform_.TransformBlock(InputBuffer_, 0, InputBlockSize_, OutputBuffer_, 0);
				TempInputBuffer_ = InputBuffer_;
				InputBuffer_ = NextInputBuffer_;
				NextInputBuffer_ = TempInputBuffer_;
				InputBufferIndex_ = NextInputBufferIndex_;
				if (localCount >= num4)
				{
					Buffer.BlockCopy(OutputBuffer_, 0, buffer, localOffset, num4);
					localOffset += num4;
					localCount -= num4;
					continue;
				}
				Buffer.BlockCopy(OutputBuffer_, 0, buffer, localOffset, localCount);
				OutputBufferIndex_ = num4 - localCount;
				Buffer.BlockCopy(OutputBuffer_, localCount, OutputBuffer_, 0, OutputBufferIndex_);
				return count;
			}
			return count;
      
		Label_0332:
			buffer3 = transform_.TransformFinalBlock(InputBuffer_, 0, InputBufferIndex_);
			OutputBuffer_ = buffer3;
			OutputBufferIndex_ = buffer3.Length;
			finalBlockTransformed_ = true;
			if (localCount < OutputBufferIndex_)
			{
				Buffer.BlockCopy(OutputBuffer_, 0, buffer, localOffset, localCount);
				OutputBufferIndex_ -= localCount;
				Buffer.BlockCopy(OutputBuffer_, localCount, OutputBuffer_, 0, OutputBufferIndex_);
				return count;
			}
			Buffer.BlockCopy(OutputBuffer_, 0, buffer, localOffset, OutputBufferIndex_);
			localCount -= OutputBufferIndex_;
			OutputBufferIndex_ = 0;
			return (count - localCount);

		}
		
		public override long Seek(long offset, SeekOrigin origin)
		{
			throw new NotSupportedException("Seek is not supported");
		}
		
		public override void SetLength(long value)
		{
			throw new NotSupportedException("SetLength is not supported");
		}
		
		public override void Write(byte[] buffer, int offset, int count)
		{
			if ( !canWrite_ ) 
			{
				throw new NotSupportedException("Cannot write to stream");
			}
			
			if (offset < 0)
			{
				throw new ArgumentOutOfRangeException("offset", "Offset < 0");
			}
			
			if (count < 0)
			{
				throw new ArgumentOutOfRangeException("count", "Count < 0");
			}
			
			if ((buffer.Length - offset) < count)
			{
				throw new ArgumentException("Offset + length out of range");
			}
			
			// If there is input in the input buffer then process it.
			if (InputBufferIndex_ > 0)
			{
				// If possible completely fill the input buffer
				if (count >= (InputBlockSize_ - InputBufferIndex_)) 
				{
					Buffer.BlockCopy(buffer, offset, InputBuffer_, InputBufferIndex_, InputBlockSize_ - InputBufferIndex_);
					offset += (InputBlockSize_ - InputBufferIndex_);
					count -= (InputBlockSize_ - InputBufferIndex_);
					InputBufferIndex_ = InputBlockSize_;
				}
				else 
				{
					// Only a partial block is possible so just update the input buffer.
					Buffer.BlockCopy(buffer, offset, InputBuffer_, InputBufferIndex_, count);
					InputBufferIndex_ += count;
					return;
				}
			}

			// Write anything left in the output buffer.
			if (OutputBufferIndex_ > 0) 
			{
				stream_.Write(OutputBuffer_, 0, OutputBufferIndex_);
				OutputBufferIndex_ = 0;
			}
			
			int transformCount;
			if (InputBufferIndex_ == InputBlockSize_) 
			{
				transformCount = transform_.TransformBlock(InputBuffer_, 0, InputBlockSize_, OutputBuffer_, 0);
				stream_.Write(OutputBuffer_, 0, transformCount);
				InputBufferIndex_ = 0;
			}
			
			if (count > 0)
			{
				while (count >= InputBlockSize_ ) 
				{
					if (transform_.CanTransformMultipleBlocks) 
					{
						int fullBlocks = count / InputBlockSize_;
						int toTransform = fullBlocks * InputBlockSize_;
						byte[] tmpBuffer = new byte[fullBlocks * OutputBlockSize_];
						transformCount = transform_.TransformBlock(buffer, offset, toTransform, tmpBuffer, 0);
						stream_.Write(tmpBuffer, 0, transformCount);
						offset += toTransform;
						count -= toTransform;
					}
					else 
					{
						transformCount = transform_.TransformBlock(buffer, offset, InputBlockSize_, OutputBuffer_, 0);
						stream_.Write(OutputBuffer_, 0, transformCount);
						offset += InputBlockSize_;
						count -= InputBlockSize_;
					}
				}
				
				// Save the remaining untransformed input in the input buffer
				Buffer.BlockCopy(buffer, offset, InputBuffer_, 0, count);
				InputBufferIndex_ += count;
			}
		}
	
		// Properties
		public override bool CanRead
		{ 
			get 
			{ 
				return canRead_;
			}
		}
		
		public override bool CanSeek 
		{ 
			get { return false; } 
		}
		
		public override bool CanWrite
		{ 
			get { return canWrite_; }
		}
		
		public override long Length 
		{
			get { throw new NotSupportedException("Length not supported"); }
		}
		
		public override long Position
		{
			get 
			{
				throw new NotSupportedException("Position not supported");
			}
			set 
			{
				throw new NotSupportedException("Position not supported");
			}
			
		}
	
		// Instance Fields
		bool canRead_;
		bool canWrite_;
		bool finalBlockTransformed_;
		Stream stream_;
		ICryptoTransform transform_;
		CryptoStreamMode mode_;

		int InputBlockSize_;
		byte[] InputBuffer_;
		int InputBufferIndex_;
		byte[] NextInputBuffer_;
		int NextInputBufferIndex_;
		int OutputBlockSize_;
		byte[] OutputBuffer_;
		int OutputBufferIndex_;
		byte[] TempInputBuffer_;
	}
	
#endif	
}
