#define BITS_12  // Size of the string table index can also be BITS_13 or BITS_14  

using System;
using System.IO;
using System.Runtime.Serialization;

namespace ICSharpCode.SharpZipLib.Src.LZW
{

	[Serializable]
	class LzwException : SharpZipBaseException
	{
		/// <summary>
		/// Deserialization constructor 
		/// </summary>
		/// <param name="info"><see cref="SerializationInfo"/> for this constructor</param>
		/// <param name="context"><see cref="StreamingContext"/> for this constructor</param>
		protected LzwException(SerializationInfo info, StreamingContext context )
			: base( info, context )
		{
		}

		public LzwException(string message) : base(message)
		{
		}
	}

	public enum StreamModeEnum {Read, Write};

	public class LzwInputStream : Stream, IDisposable
	{
		const int INIT_BITS = 9;              // Initial number of bits per code

		const int BIT_MASK  =  0x1f;           // Mask for 'number of compression bits'
		/* Mask 0x20 is reserved to mean a fourth header byte, and 0x40 is free.
		 * It's a pity that old uncompress does not check bit 0x20. That makes
		 * extension of the format actually undesirable because old compress
		 * would just crash on the new format instead of giving a meaningful
		 * error message. It does check the number of bits, but it's more
		 * helpful to say "unsupported format, get a new version" than
		 * "can only handle 16 bits".
		 */

		const int BLOCK_MODE = 0x80;
		/* Block compression: if table is full and compression rate is dropping,
		 * clear the dictionary.
		 */

		const int LZW_RESERVED = 0x60; // reserved bits

		const int CLEAR = 256;       // flush the dictionary
		const int FIRST = (CLEAR+1); // first free entry

		/// <param name="stream">Underlying stream</param>
		/// <param name="streamMode">Reading or writing the stream</param>
		/// <param name="notificationByteCount">Number of bytes before an event is raised</param>
		public LzwInputStream(Stream stream)
		{
			Init(stream);
		}

		/// <summary>
		/// Initialize this instance.
		private void Init(Stream stream)
		{
			stream_ = stream;

			if (!stream.CanRead)
			{
				throw new LzwException("Underlying stream does not support reading");
			}

			ReadHeader();
		}
	
		int MAXCODE(int n)
		{
			return 1 << n;
		}

		void ReadHeader()
		{
			int nextByte = stream_.ReadByte();
			if ( nextByte != 0x1F )
			{
				throw new LzwException("Bad magic");
			}
         
			nextByte = stream_.ReadByte();
			if ( nextByte != 0x9D )
			{
				throw new LzwException("Bad magic");
			}

			int maxbits = stream_.ReadByte();
			if ( nextByte < 0 )
			{
				throw new EndOfStreamException("Unexpected end of stream");
			}

			block_mode = (maxbits & BLOCK_MODE) != 0;
			if ((maxbits & LZW_RESERVED) != 0) 
			{
				// Unknown flags... leave for now.
			}

			maxbits &= BIT_MASK;
			int maxmaxcode = MAXCODE(maxbits);
    
			if (maxbits > BITS) 
			{
				throw new LzwException(string.Format("Cannot decompress {0} bits", maxbits));
			}


			nextCode = block_mode ? CLEAR : FIRST;

			prefixCode = new int[TABLE_SIZE];
			codeValue = new int[TABLE_SIZE];
			appendCharacter = new byte[TABLE_SIZE];
			decodeStack = new byte[4096];

			nextCode = 256;				
			for (int i = 0; i < TABLE_SIZE; ++i)
				codeValue[i]=-1;
		}

		/// <summary>
		/// Get value indicating if stream can be read.
		/// </summary>
		public override bool CanRead
		{
			get
			{
				return true;
			}
		}

		/// <summary>
		/// Get a value indicating is this stream can be written.
		/// </summary>
		public override bool CanWrite		
		{
			get
			{
				return false;
			}
		}

		/// <summary>
		/// This call will always return false because you can not seek 
		/// while expanding.
		/// </summary>
		public override bool CanSeek		
		{
			get
			{
				return false;
			}
		}


		/// <summary>
		/// The call to dispose will close the stream.
		/// </summary>
		public void Dispose()
		{
			if (disposed)
			{
				throw new ObjectDisposedException("The stream has already been disposed (closed)");
			}
			disposed = true;
  
			Flush();
			stream_.Close();
		}

		/// <summary>
		/// Close the stream after the expansion has been completed.
		/// </summary>
		public override void Close()
		{
			Dispose();
		}


		/// <summary>
		/// Flush will flush the underlying stream.
		/// </summary>
		public override void Flush()		
		{
			stream_.Flush();
		}

		/// <summary>
		/// Length can not return the length of the expanded stream because
		/// this value is unknown until the end of the underlying stream
		/// has been reached. Therefore it returns the number of so far
		/// processed bytes.
		/// </summary>
		public override long Length
		{
			get
			{
				return processedBytes;
			}
		}

		/// <summary>
		/// Current position in the underlying stream.
		/// </summary>
		public override long Position
		{
			get
			{
				return stream_.Position;
			}
			set
			{
				throw new NotSupportedException("Cannot set position");
			}
		}

		public override long Seek(long offset, SeekOrigin loc)
		{
			throw new NotSupportedException("Cannot seek");
		}

		public override void SetLength(long value)
		{
			throw new NotSupportedException("Cannot set length");
		}

		/// <summary>
		/// Read will process a compressed input stream and insert the expanded
		/// bytes into the buffer of bytes. The number of bytes inserted into
		/// the buffer will certainly exceed the number of bytes read from the
		/// input buffer depending on the compression ratio. 
		/// </summary>
		/// <param name="buffer">Buffer for expanded bytes</param>
		/// <param name="offset">Starting position to insert bytes</param>
		/// <param name="count">Number of bytes to be inserted</param>
		/// <returns>Number of bytes inserted into the buffer</returns>
		public override int Read(byte[] buffer, int offset, int count)
		{
			// If count is 0 we are done
			if (count == 0)
				return 0;

			// If we have reached the end of the input stream before already
			if (gameOver)
				return 0;

			int bytePosition = 0;

			// May be there are more bytes on the stack left from the last
			// call to Read. We need to process all those bytes until 
			// the stack is empty before proceeding the expansion of 
			// more codes from the input stream.
			if (stringPointer >= decodeStackPointer)
			{
				// While more known bytes are on the stack
				while (stringPointer >= decodeStackPointer)
				{
					buffer[offset + bytePosition] = decodeStack[stringPointer--];
					bytePosition++;
					processedBytes++;
					if (stringPointer >= decodeStackPointer && bytePosition == count)
						return bytePosition;
				}

				if (nextCode <= MAX_CODE)
				{
					prefixCode[nextCode] = oldCode;
					appendCharacter[nextCode] = character;
					nextCode++;
				}
				oldCode = newCode;
			}

			// Initialize oldCode and character before starting the algorithm
			if (firstByte)
			{
				oldCode = getCode();
				character = (byte)oldCode;    
				buffer[offset+ bytePosition] = character;
				bytePosition++;
				firstByte = false;
			}

			// Expand codes from the input stream until the end of the compressed
			// stream has been reached (MAX_VALUE)
			while (bytePosition < count && (newCode = getCode()) != MAX_VALUE)
			{

				if (newCode >= nextCode)
				{
					decodeStack[decodeStackPointer] = character;
					stringPointer = decodeString(decodeStackPointer+1, oldCode);
				}
				else 	// May be there is a matching pattern
				{
					stringPointer = decodeString(decodeStackPointer, newCode);
				}

				character = decodeStack[stringPointer];
				while (stringPointer >= decodeStackPointer)
				{
					buffer[offset + bytePosition] = decodeStack[stringPointer--];
					bytePosition++;
					processedBytes++;
					if (stringPointer >= decodeStackPointer && bytePosition == count)
						return bytePosition;
				}

				if (nextCode <= MAX_CODE)
				{
					prefixCode[nextCode] = oldCode;
					appendCharacter[nextCode] = character;
					nextCode++;
				}
				oldCode = newCode;

			}

			if (newCode == MAX_VALUE)
				gameOver = true;

			return bytePosition;

		}


		/// <summary>
		/// Decode the bytes from the hash list, where known byte patterns
		/// are maintained. It will transfer those expanded bytes to a stack where
		/// they are ready for Read or Write.
		/// </summary>
		/// <param name="bufferPointer">Pointer to the last stack position</param>
		/// <param name="code">Pointer to the chain of known bytes</param>
		/// <returns>Pointer to the last stack position</returns>
		private int decodeString(int bufferPointer, int code)
		{
			int i = 0;

			while (code > 255)
			{
				decodeStack[bufferPointer++] = appendCharacter[code];
				code = prefixCode[code]; 

				// Check error in compressed bytes
				if (i++ >= 4094)
				{
					throw new LzwException("Fatal error during code expansion.");
				}

			}

			decodeStack[bufferPointer] = (byte)code;

			return bufferPointer;
		} // decodeString	

		/// <summary>
		/// This function will get the next code (12, 13 or 14 bit value) from
		/// the compressed stream. It needs to combine two 8 bit bytes to get a 
		/// complete code.
		/// </summary>
		/// <returns>Next code from the compressed stream.</returns>
		private int getCode()
		{
			int returnValue;

			while (inputBitCount <= 24)
			{
				int temp = stream_.ReadByte(); // -1 marks the end of stream
				if (temp == -1)
					temp = 0;

				byteCount++;
				inputBitBuffer |= (ulong) temp << (24-inputBitCount);
				inputBitCount += 8;
			}
			returnValue = (int)((inputBitBuffer >> (32-BITS)) & MAX_VALUE);
			inputBitBuffer <<= BITS;
			inputBitCount -= BITS;

			return returnValue;
		}


		public override void Write(byte[] buffer, int offset, int count)
		{
			throw new NotSupportedException("Writing not supported by this stream");
		}
		
		
		bool block_mode;

#if BITS_14
		const int BITS = 14;
		const int TABLE_SIZE = 18041;        
#endif
#if BITS_13                   
		const int BITS = 13;
		const int TABLE_SIZE = 9029;
#endif
#if BITS_12
		const int BITS = 12;
		const int TABLE_SIZE = 5021;
#endif

		const int HASHING_SHIFT = BITS - 8;
		const int MAX_VALUE = (1 << BITS) -1;
		const int MAX_CODE = MAX_VALUE - 1;

		int[] prefixCode;
		int[] codeValue;

		byte[] appendCharacter;
		byte[] decodeStack;

		int nextCode = 0;
		bool disposed = false;
		Stream stream_;

		int inputBitCount = 0;
		ulong inputBitBuffer = 0L;

		byte[] buffer;
		int offset = 0;
		int count;
		int byteCount = 0;
		int newCode = 0;
		int oldCode = 0;
		int decodeStackPointer = 0;
		byte character = 0;
		int stringPointer = -1;
		bool firstByte = true;
		long processedBytes = 0;
		bool gameOver = false;
	}
}
